@@ -1,27 +1,32 @@
+/* xa65 - 65xx/65816 cross-assembler and utility suite
+ *
+ * Copyright (C) 1989-1997 Andr Fachat (a.fachat@physik.tu-chemnitz.de)
+ * maintained by Cameron Kaiser
+ *
+ * Core tokenizing module/pass 1 and pass 2
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ */
 
-/*
-    xa65 - 6502 cross assembler and utility suite
-    Copyright (C) 1989-1998 Andr Fachat (a.fachat@physik.tu-chemnitz.de)
-
-    This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-*/
-
+/* enable this to turn on (copious) optimization output */
+/* #define DEBUG_AM */
 
 #include <ctype.h>
 #include <stdio.h>
 
+#include "xad.h"
 #include "xah.h"
 #include "xah2.h"
 
@@ -43,65 +48,42 @@ static void tg_hex(signed char*,int*,int*);
 static void tg_oct(signed char*,int*,int*);
 static void tg_bin(signed char*,int*,int*);
 
+/* assembly mnemonics and pseudo-op tokens */
+/* ina and dea don't work yet */
 static char *kt[] ={ 
-     "adc","and","asl","bbr","bbs","bcc","bcs","beq",
-     "bit","bmi",
-     "bne","bpl","bra","brk","bvc","bvs",
+/*    1     2     3    4      5     6      7   8      9     10   */
+     "adc","and","asl","bbr","bbs","bcc","bcs","beq","bit","bmi",
+     "bne","bpl","bra","brk","bvc","bvs","brl","clc","cld","cli",
+/*
+     "clv","cmp","cpx","cpy","cop","dea","dec","dex","dey","eor",
+*/
+     "clv","cmp","cpx","cpy","cop",/*"dea",*/"dec","dex","dey","eor",
 
-     /* new */
-     "brl",
+/*
+     "ina","inc","inx","iny","jmp","jsr","lda","ldx","ldy","lsr",
+*/
+     /*"ina",*/"inc","inx","iny","jmp","jsr","lda","ldx","ldy","lsr",
+     "mvp","mvn","nop","ora","pha","php","phx","phy","pla","plp",
+     "plx","ply","phb","phd","phk","plb","pld","pea","pei","per",
 
-     "clc","cld","cli","clv","cmp","cpx","cpy",
+     "rmb","rol","ror","rti","rts","rep","rtl","sbc","sec","sed",
+     "sei","smb","sta","stx","sty","stz","sep","stp","tax","tay",
+     "trb","tsb","tsx","txa","txs","tya","txy","tyx","tcd","tdc",
 
-     /* new */
-     "cop",
+     "tcs","tsc","wai","wdb","xba","xce", /* 96 */
 
-     "dec","dex","dey","eor",
-     "inc","inx","iny","jmp","jsr","lda","ldx","ldy",
-     "lsr",
-
-     /* new */
-     "mvp","mvn",
-
-     "nop","ora","pha","php","phx","phy","pla",
-     "plp","plx","ply",
-
-     /* new */
-     "phb","phd","phk","plb","pld",
-     "pea","pei","per",
-
-     "rmb","rol",
-     "ror","rti","rts",
-
-     /* new */
-     "rep","rtl",
-
-     "sbc","sec","sed","sei","smb","sta",
-     "stx","sty","stz",
-
-     /* new */
-     "sep","stp",
-
-     "tax","tay","trb","tsb","tsx","txa","txs","tya",
-
-     /* new */
-     "txy","tyx","tcd","tdc","tcs","tsc",
-     "wai","wdb",
-     "xba","xce",
-
-     ".byt",".word",".asc",".dsb",
-     ".(",".)","*=", ".text", ".data", ".bss", ".zero",
-     ".fopt", ".byte", ".end", ".list", ".xlist", 
-     ".dupb", ".blkb", ".db", ".dw", ".align",
-     ".block", ".bend",
-
-     ".al",".as",".xl",".xs"
+     ".byt",".word",".asc",".dsb", ".(", ".)", "*=", ".text",".data",".bss",
+     ".zero",".fopt", ".byte", ".end", ".list", ".xlist", ".dupb", ".blkb", ".db", ".dw",
+     ".align",".block", ".bend",".al",".as",".xl",".xs" /* 96 + 27 = 123 */
 
 };
 
 static int lp[]= { 0,1,1,1,1,2,2,1,1,1,2,2,2,1,1,1,2,2 };
 
+/* index into token array for pseudo-ops */
+/* last valid mnemonic */
 #define   Lastbef   93
+/* last valid token+1 */
 #define   Anzkey    121
 
 #define   Kbyt      Lastbef+1
@@ -141,6 +123,41 @@ static int ktp[]={ 0,3,17,25,28,29,29,29,29,32,34,34,38,40,41,42,58,
 
 #define   Admodes   24
 
+/* 
+ * opcodes for each addressing mode
+ * high byte: supported architecture (no bits = original NMOS 6502)
+ *  bit 1: R65C02
+ *  bit 2: 65816
+ *  bit 3: 65816 and allows 16-bit quantity (immediate only)
+ * low byte: opcode itself
+ *
+ * each opcode is indexed in this order: *=65816, ^=R65C02
+ * 00 = implied
+ * 01 = zero page
+ * 02 = zero page,x
+ * 03 = direct page,y*
+ * 04 = direct page (indirect)*
+ * 05 = (indirect,x)
+ * 06 = (indirect),y
+ * 07 = immediate (8-bit)
+ * 08 = absolute
+ * 09 = absolute,x
+ * 10 = absolute,y
+ * 11 = relative
+ * 12 = (indirect-16) i.e., jmp (some_vector)
+ * 13 = (absolute,x)*
+ * 14 = zero page+relative test'n'branch ^
+ * 15 = zero page clear'n'set'bit ^
+ * 16 = relative long*
+ * 17 = absolute long*
+ * 18 = absolute long,x*
+ * 19 = stack relative*
+ * 20 = stack relative (indirect),y*
+ * 21 = direct page (indirect long)*
+ * 22 = direct page (indirect long),y*
+ * 23 = (indirect long)
+ */
+
 static int ct[Lastbef+1][Admodes] ={
 /*     0    1    2    3    4    5    6    7    8    9   10   11   12   13   14   15   16   17   18   19   20   21   22   23   imm */
 {     -1,  0x65,0x75,-1,0x172,0x61,0x71,0x469,0x6d,0x7d,0x79,-1,  -1,  -1,  -1,  -1,  -1,0x26f,0x27f,0x263,0x273,0x267,0x277,-1 },  /*adc*/
@@ -159,9 +176,7 @@ static int ct[Lastbef+1][Admodes] ={
 {     0x00,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*brk*/
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  0x50,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*bvc*/
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  0x70,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*bvs*/
-
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,0x282, -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*brl*/
-
 {     0x18,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*clc*/
 {     0xd8,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*cld*/
 {     0x58,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*cli*/
@@ -169,13 +184,15 @@ static int ct[Lastbef+1][Admodes] ={
 {     -1,  0xc5,0xd5,-1, 0x1d2,0xc1,0xd1,0x4c9,0xcd,0xdd,0xd9,-1, -1,  -1,  -1,  -1,-1,0x2cf,0x2df,0x2c3,0x2d3,0x2c7,0x2d7,-1 },  /*cmp*/
 {     -1,  0xe4,-1,  -1,  -1,  -1,  -1, 0x8e0,0xec,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*cpx*/
 {     -1,  0xc4,-1,  -1,  -1,  -1,  -1, 0x8c0,0xcc,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*cpy*/
-
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1,  0x202,-1,  -1,  -1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*cop*/
-
+/*
+{     0x13a,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },*/ /*dea*/
 {     0x13a,0xc6,0xd6,-1,  -1,  -1,  -1,  -1,  0xce,0xde,-1,  -1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*dec*/
 {     0xca,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*dex*/
 {     0x88,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*dey*/
 {     -1,  0x45,0x55,-1, 0x152,0x41,0x51,0x449,0x4d,0x5d,0x59,-1, -1,  -1,  -1,  -1,-1,0x24f,0x25f,0x243,0x253,0x247,0x257,-1 },  /*eor*/
+/*
+{     0x11a,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },*/ /*ina*/
 {     0x11a,0xe6,0xf6,-1,  -1,  -1,  -1,  -1,  0xee,0xfe,-1,  -1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*inc*/
 {     0xe8,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*inx*/
 {     0xc8,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*iny*/
@@ -185,10 +202,8 @@ static int ct[Lastbef+1][Admodes] ={
 {     -1,  0xa6,-1,  0xb6,-1,  -1,  -1, 0x8a2,0xae,-1,  0xbe,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*ldx*/
 {     -1,  0xa4,0xb4,-1,  -1,  -1,  -1, 0x8a0,0xac,0xbc,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*ldy*/
 {     0x4a,0x46,0x56,-1,  -1,  -1,  -1,  -1,  0x4e,0x5e,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*lsr*/
-
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  0x244,-1,  -1,  -1 ,-1,  -1,  -1,  -1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*mvp*/
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  0x254,-1,  -1,  -1 ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*mvn*/
-
 {     0xea,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*nop*/
 {     -1,  0x05,0x15,-1, 0x112,0x01,0x11,0x409,0x0d,0x1d,0x19,-1, -1,  -1,  -1,  -1,-1,0x20f,0x21f,0x203,0x213,0x207,0x217,-1 },  /*ora*/
 {     0x48,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*pha*/
@@ -199,7 +214,6 @@ static int ct[Lastbef+1][Admodes] ={
 {     0x28,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*plp*/
 {     0x1fa,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*plx*/
 {     0x17a,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*ply*/
-
 {     0x28b,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*phb*/
 {     0x20b,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*phd*/
 {     0x24b,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*phk*/
@@ -208,16 +222,13 @@ static int ct[Lastbef+1][Admodes] ={
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  0x2f4,-1,  -1, -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*pea*/
 {     -1,  -1,  -1,  -1,0x2d4, -1,  -1,  -1,  -1,  -1,  -1,  -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*pei*/
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1  ,-1,  -1,  -1,  -1, 0x262,-1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*per*/
-
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,0x107, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*rmb*/
 {     0x2a,0x26,0x36,-1,  -1,  -1,  -1,  -1,  0x2e,0x3e,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*rol*/
 {     0x6a,0x66,0x76,-1,  -1,  -1,  -1,  -1,  0x6e,0x7e,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*ror*/
 {     0x40,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*rti*/
 {     0x60,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*rts*/
-
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1, 0x2c2,-1,  -1,  -1,  -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*rep*/
 {     0x26b,-1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*rtl*/
-
 {     -1,  0xe5,0xf5,-1, 0x1f2,0xe1,0xf1,0x4e9,0xed,0xfd,0xf9,-1, -1,  -1,  -1,  -1,-1,0x2ef,0x2ff,0x2e3,0x2f3,0x2e7,0x2f7,-1 },  /*sbc*/
 {     0x38,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*sec*/
 {     0xf8,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*sed*/
@@ -227,10 +238,8 @@ static int ct[Lastbef+1][Admodes] ={
 {     -1,  0x86,-1,  0x96,-1,  -1,  -1,  -1,  0x8e,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*stx*/
 {     -1,  0x84,0x94,-1,  -1,  -1,  -1,  -1,  0x8c,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*sty*/
 {     -1,  0x164,0x174,-1, -1,  -1,  -1,  -1, 0x19c,0x19e,-1, -1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*stz*/
-
 {     -1,  -1,  -1,  -1,  -1,  -1,  -1, 0x2e2,-1,  -1,  -1,  -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*sep*/
 {     0x2db,-1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*stp*/
-
 {     0xaa,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*tax*/
 {     0xa8,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*tay*/
 {     -1,  0x114,-1,  -1,  -1,  -1,  -1, -1, 0x11c,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*trb*/
@@ -239,7 +248,6 @@ static int ct[Lastbef+1][Admodes] ={
 {     0x8a,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*txa*/
 {     0x9a,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*txs*/
 {     0x98,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*tya*/
-
 {     0x29b,-1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*txy*/
 {     0x2bb,-1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*tyx*/
 {     0x25b,-1, -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1  ,-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 },  /*tcd*/
@@ -254,39 +262,38 @@ static int ct[Lastbef+1][Admodes] ={
 
 } ;
 
-#if 0
-static int cf[] ={ 0,0,0,1,1,0,0,0,0,0,0,0,1,  /* 1= CMOS     */
-            0,0,0,0,0,0,0,0,0,0,
-            0,0,0,0,0,0,0,0,0,0,
-            0,0,0,0,0,0,0,1,1,0,0,1,1,1,
-            0,0,0,0,0,0,0,0,1,0,0,0,
-            1,0,0,1,1,0,0,0,0 };
-#endif
-
 #define   Syntax    14
 #define   Maxbyt    4
 
+/* grouped syntaxes. each row = operand type, column = bytes allowed */
 static int at[Syntax][Maxbyt] ={
-{     0,   -1,  -1  ,-1 },
-{     -1,  7,   -1  ,-1 },
-{     -1,  15,  -1  ,-1 },
-{     -1,  -1,  14  ,-1 },
-{     -1,  1,    8  ,17 },
-{     -1,  2,    9  ,18 },
-{     -1,  3,   10  ,-1 },
-{     -1,  4,   12  ,-1 },
-{     -1,  5,   13  ,-1 },
-{     -1,  6,   -1  ,-1 },
-
-{     -1,  21,  23  ,-1 },
-{     -1,  22,  -1  ,-1 },
-{     -1,  19,  -1  ,-1 },
-{     -1,  20,  -1  ,-1 }
+{     0,   -1,  -1  ,-1 },	/* implied: no operand */
+{     -1,  7,   -1  ,-1 },	/* immediate: single byte operand only */
+{     -1,  15,  -1  ,-1 },	/* relative: single byte operand only */
+{     -1,  -1,  14  ,-1 },	/* test'n'branch: two bytes only */
+{     -1,  1,    8  ,17 },	/* addressing: 1 byte for zp,
+					2 for absolute,
+					3 for absolute long */
+{     -1,  2,    9  ,18 },	/* ,x: same */
+{     -1,  3,   10  ,-1 },	/* ,y: 1 byte for dp,y,
+					2 for absolute,y */
+{     -1,  4,   12  ,-1 },	/* (indirect): 1 byte for (dp),
+					2 for (absolute) */
+{     -1,  5,   13  ,-1 },	/* (i,x): 1 byte for (zp,x),
+					2 for (a,x) */
+{     -1,  6,   -1  ,-1 },	/* (i),y: 1 byte only */
+{     -1,  21,  23  ,-1 },	/* (indirect long): 1 byte for (dp),
+					2 for (a) */
+{     -1,  22,  -1  ,-1 },	/* (indirect long),y: 1 byte only */
+{     -1,  19,  -1  ,-1 },	/* stack relative: 1 byte only */
+{     -1,  20,  -1  ,-1 }	/* SR (in),y: 1 byte only */
 };
 
-
 #define   AnzAlt    5
 
+/* disambiguation table. for example, arbitrary instruction xxx $0000 could
+	be either interpreted as an absolute operand, or possibly relative.
+	note: does not look at comma or after, if present. */
 static int xt[AnzAlt][2] ={ /* Alternativ Adr-Modes  */
 {     8,   11 },      /* abs -> rel  */
 {     2,   3  },      /* z,x -> z,y  */
@@ -296,27 +303,37 @@ static int xt[AnzAlt][2] ={ /* Alternativ Adr-Modes  */
 {     8,   16 }       /* abs -> relong */
 };
 
+/* cross check: instruction should be this many bytes long in total */
+/* indexed by addressing mode */
 static int le[] ={ 1,2,2,2,2,2,2,2,3,3,3,2,3,3,3,2,
                 /* new modes */ 3,4,4,2,2,2,2,3 };
 
+/* indicates absolute->zp optimizable addressing modes (abs->zp) */
+/* indexed by addressing mode */
 static int opt[] ={ -1,-1,-1,-1,-1,-1,-1,-1,1,2,3,-1,4,5,-1,-1,
            /*new*/ -1,8,9,-1,-1,-1,-1,-1 }; /* abs -> zp */
 
-#define   cval(s)   256*((s)[1]&255)+((s)[0]&255)
-#define   wval(i,v) t[i++]=T_VALUE;t[i++]=(v)&255;t[i++]=((v)>>8)&255;t[i++]=((v)>>16)&255;
-
 /* pass 1 */
 int t_p1(signed char *s, signed char *t, int *ll, int *al)
 {
      static int er,l,n,v,nk,na1,na2,bl,am,sy,i,label,byte; /*,j,v2 ;*/
      int afl = 0;
 
+/* notes and typical conventions ... er = error code
+	am = addressing mode in use
+*/
+
      bl=0;
      *al = 0;
 
 /*     printf("\n"); */
 
+     /* convert the next token from string s */
+#ifdef DEBUG_AM
+fprintf(stderr, "- p1 %d starting -\n", pc[segment]);
+#endif
      er=t_conv(s,t,&l,pc[segment],&nk,&na1,&na2,0,&byte);
+     /* leaving our token sequence in t */
 
      *ll=l;
 /*
@@ -327,17 +344,26 @@ int t_p1(signed char *s, signed char *t, int *ll, int *al)
 */
      /* if text/data produced, then no more fopt allowed in romable mode */
      if((romable>1) && (t[0]<Kopen || t[0]==Kbyte || t[0]==Kpcdef)) {
-       afile->base[SEG_TEXT] = pc[SEG_TEXT] = romadr + h_length();
+       afile->base[SEG_TEXT] = pc[SEG_TEXT] = romaddr + h_length();
        romable=1;
      }
 
      if(!er)
      {
+
+/*
+ *
+ * pseudo-op dispatch (except .byt, .asc)
+ *
+ */
           n=t[0];
 	  /* TODO: make that a big switch statement... */
+	  /* maybe later. Cameron */
+
 	  if(n==Kend || n==Klist || n==Kxlist) {
 	    *ll = 0;		/* ignore */
 	  } else
+
 	  if(n==Kfopt) {
 	    if(romable==1) er=E_ROMOPT;
 	    t[0] = Kbyt;
@@ -434,7 +460,7 @@ printf(" wrote %02x %02x %02x %02x %02x %02x\n",
 	       dsb_len = 0;
           } else
 	  if(n==Ktext) {
-/*	    if(segment!=SEG_ABS) {*/
+/*	    if(segment!=SEG_ABS) {    */
 	      segment = relmode ? SEG_TEXT : SEG_ABS;
 	      t[0]=Ksegment;
 	      t[1]=segment;
@@ -442,10 +468,10 @@ printf(" wrote %02x %02x %02x %02x %02x %02x\n",
               er=E_OKDEF;
 /*	    } else {
 	      er=E_ILLSEGMENT;
-	    }*/
+	    }                        */
 	  } else
 	  if(n==Kdata) {
-/*  	    if(segment!=SEG_ABS) {*/
+/*  	    if(segment!=SEG_ABS) {   */
 	      segment = SEG_DATA;
 	      t[0]=Ksegment;
 	      t[1]=SEG_DATA;
@@ -453,10 +479,10 @@ printf(" wrote %02x %02x %02x %02x %02x %02x\n",
               er=E_OKDEF;
 /*	    } else {
 	      er=E_ILLSEGMENT;
-	    }*/
+	    }                        */
 	  } else
 	  if(n==Kbss) {
-/*  	    if(segment!=SEG_ABS) {*/
+/*  	    if(segment!=SEG_ABS) {   */
 	      segment = SEG_BSS;
 	      t[0]=Ksegment;
 	      t[1]=SEG_BSS;
@@ -464,10 +490,10 @@ printf(" wrote %02x %02x %02x %02x %02x %02x\n",
               er=E_OKDEF;
 /*	    } else {
 	      er=E_ILLSEGMENT;
-	    }*/
+	    }                        */
 	  } else
 	  if(n==Kzero) {
-/*  	    if(segment!=SEG_ABS) {*/
+/*  	    if(segment!=SEG_ABS) {   */
 	      segment = SEG_ZERO;
 	      t[0]=Ksegment;
 	      t[1]=SEG_ZERO;
@@ -475,7 +501,7 @@ printf(" wrote %02x %02x %02x %02x %02x %02x\n",
               er=E_OKDEF;
 /*	    } else {
 	      er=E_ILLSEGMENT;
-	    }*/
+	    }                        */
 	  } else
 	  if(n==Kalign) {
 	    int tmp;
@@ -506,48 +532,71 @@ printf(" wrote %02x %02x %02x %02x %02x %02x\n",
 	      er=E_ILLSEGMENT;
 	    }
 	  } else
-               er=t_p2(t,ll,(byte==2)?0:1, al);
+/* optimization okay on pass 1: use 0 for fl */
+		{
+#ifdef DEBUG_AM
+fprintf(stderr, "E_OK ... t_p2 xat.c\n");
+#endif
+               er=t_p2(t,ll,(0 | byte), al);
+	}
           
      } else
-     if(er==E_NODEF)
+     if(er==E_NODEF) /* no label was found from t_conv! */
      {
-          er=E_OK;
-          n=t[0];
+	er = E_OK;
+          n=t[0];	/* look at first token */
+
+/* mnemonic dispatch -- abbreviated form in t_p2, but changed here to not
+do anything other than 24-bit optimization since we don't know the value of
+the label */
+
+		/* choose addressing mode; add commas found */
 
           if(n>=0 && n<=Lastbef)
           {
                if(t[1]==T_END)
                {
-                    sy=0;
+                    sy=0;	/* implied */
                } else
                if(t[1]=='#')
                {
-                    sy=1+nk;
+                    sy=1+nk;	/* immediate */
                } else
                if(t[1]=='(')
                {
-                    sy=7+nk;
+                    sy=7+nk;	/* computed */
                } else
-                    sy=4+nk;
+                    sy=4+nk;	/* absolute or zero page */
 
+		/* length counter set to maximum length + 1 */
                bl=Maxbyt+1;
                
+		/* find best fit for length of this operand */
                while(--bl)
                {
+
+		/* look at syntax table (at) using syntax (sy) as index.
+			is there an addressing mode for an operand
+			of this length? am = addressing mode */
+
                     if((am=at[sy][bl-1])>=0)
                     {
-                         if(am>Admodes-1)
+                         if(am>Admodes-1) /* no, it's -1, syntax error */
                          {
                               er=E_SYNTAX;
                               break;
                          }
-                         if(ct[n][am]>=0)
+                         if(ct[n][am]>=0) /* yes, valid token *and* mode,
+						so we're done */
                               break;
 
+			/* no valid mode for this token, see if it's something
+				ambiguous; if so, try to interpret in that
+				context. */
                          for(v=0;v<AnzAlt;v++)
                               if(xt[v][0]==am && ct[n][xt[v][1]]>=0)
                                    break;
-                         if(v<AnzAlt) 
+                         if(v<AnzAlt) /* got a match for another context */
                          {
                               am=xt[v][1];
                               break;
@@ -555,21 +604,41 @@ printf(" wrote %02x %02x %02x %02x %02x %02x\n",
                     }
                }
 
-               if (t[l-1]!='@')
+		/* optimize operand length for 24-bit quantities */
+		/* look at cast byte from t_conv */
+               if (t[l-1]!='@' && t[l-1] != '!')
                {
-                     if(bl && !er && opt[am]>=0 && am>16)
+                     if(bl && !er && opt[am]>=0 && am>16) /* <<< NOTE! */
                           if(ct[n][opt[am]]>=0)
                                am=opt[am];
                }
+		/* if ` is declared, force further optimization */
+		if (t[l-1]=='`') {
+			if (opt[am]<0 || ct[n][opt[am]]<0)
+				errout(E_ADRESS);
+			am=opt[am];
+		}
+			
 
+
+		/* couldn't match anything for this opcode */
                if(!bl)
                     er=E_SYNTAX;
                else {
                     bl=le[am];
-                    if( ((ct[n][am]&0x400) && memode) || ((ct[n][am]&0x800) && xmode))
+                    if( ((ct[n][am]&0x400) && memode) ||
+			((ct[n][am]&0x800) && xmode)) {
                           bl++;
+			}
                 }
+		if (er == E_NODEF)
+			er = E_OK;
 
+/*
+ *
+ * .byt, .asc dispatch
+ *
+ */
 
           } else
           if(n==Kbyt || n==Kasc)
@@ -612,12 +681,19 @@ printf(" wrote %02x %02x %02x %02x %02x %02x\n",
      return(er);
 }
 
+/*t_pass 2*/
 int t_p2(signed char *t, int *ll, int fl, int *al)
 {
      static int afl,nafl, i,j,k,er,v,n,l,bl,sy,am,c,vv[3],v2,label;
      static int rlt[3];	/* relocation table */
      static int lab[3];	/* undef. label table */
 
+#if(0)
+     (void)fl;		/* quench warning */
+#endif
+/* fl was not used in 2.2.0 so I'm overloading it for zp-optimization
+     control */
+
      er=E_OK;
      bl=0;
      if(*ll<0) /* <0 bei E_OK, >0 bei E_OKDEF     */
@@ -999,17 +1075,53 @@ int t_p2(signed char *t, int *ll, int fl, int *al)
                     }
                }
 
+/* FIXIT1 
+		fprintf(stderr, "t has: ");
+		for(v=0;v<l;v++) {
+			fprintf(stderr, "%02x ", t[v]);
+		}
+		fprintf(stderr, "\n");
+*/
+
+/* only do optimization if we're being called in pass 1 -- never pass 2 */
+/* look at cast byte */
                if (t[*ll-1]!='@')
                {
-                    if(bl && !er && !(vv[0]&0xff0000) && opt[am]>=0 && am>16)
+#ifdef DEBUG_AM
+fprintf(stderr,
+"b4: pc= %d, am = %d and vv[0] = %d, optimize = %d, bitmask = %d\n",
+	pc[segment], am, vv[0], fl, (vv[0]&0xffff00));
+#endif
+
+/* terrible KLUDGE!!!! OH NOES!!!1!
+   due to the way this is constructed, you must absolutely always specify @ to
+   get an absolute long or it will absolutely always be optimized down */
+
+                    if(bl && am>16 &&
+			!er && !(vv[0]&0xff0000) && opt[am]>=0)
                          if(ct[n][opt[am]]>=0)
                               am=opt[am];
-
+#ifdef DEBUG_AM
+fprintf(stderr,
+"aftaa1: pc= %d, am = %d and vv[0] = %d, optimize = %d, bitmask = %d\n",
+	pc[segment], am, vv[0], fl, (vv[0]&0xffff00));
+#endif
                     if(t[*ll-1]!='!') {
-                         if(bl && !er && !(vv[0]&0xffff00) && opt[am]>=0)
-                              if(ct[n][opt[am]]>=0)
-                                   am=opt[am];
+                         if(bl && !er && !(vv[0]&0xffff00) && opt[am]>=0) {
+                              if(ct[n][opt[am]]>=0) {
+				if (!fl || t[*ll-1]=='`') {
+                                   	am=opt[am];
+				} else {
+					errout(W_FORLAB);
+				}
+			      }
+			}
                     }
+#ifdef DEBUG_AM
+fprintf(stderr,
+"aftaa2: pc=%d, am=%d and vv[0]=%d, optimize=%d, bitmask=%d, op=%d\n",
+	pc[segment], am, vv[0], fl, (vv[0]&0xffff00), ct[n][opt[am]]);
+#endif
                }
 
                if(!bl)
@@ -1017,12 +1129,16 @@ int t_p2(signed char *t, int *ll, int fl, int *al)
                else
                {
                     bl=le[am];
-                    if( ((ct[n][am]&0x400) && memode) || ((ct[n][am]&0x800) && xmode))
+                    if( ((ct[n][am]&0x400) && memode) || ((ct[n][am]&0x800) && xmode)) {
                          bl++;
+		}
                     *ll=bl;
 
                }
 
+#ifdef DEBUG_AM
+fprintf(stderr, "byte length is now %d\n", bl);
+#endif
 
                if(!er)
                {
@@ -1043,8 +1159,12 @@ int t_p2(signed char *t, int *ll, int fl, int *al)
                     {
                          if((am<8 && !( ((ct[n][am]&0x400) && memode) || ((ct[n][am]&0x800) && xmode) )) || (am>=19 && am!=23))
                          {
-                              if(vv[0]&0xff00)
+                              if(vv[0]&0xff00) {
+#ifdef DEBUG_AM
+fprintf(stderr, "address mode: %i address: %i\n", am, vv[0]);
+#endif
                                    er=E_OVERFLOW;
+				}
                               else
                                    t[1]=vv[0];
 /*if(rlt[0]) printf("relocation 1 byte %04x at pc=$%04x, value now =$%04x\n",rlt[0],pc[segment]+1,*vv); */
@@ -1052,8 +1172,12 @@ int t_p2(signed char *t, int *ll, int fl, int *al)
                          } else
                          if(((am<14 || am==23) && am!=11) || am==7)
                          {
-                              if (vv[0]>0xffff)
+                              if (vv[0]>0xffff) {
+#ifdef DEBUG_AM
+fprintf(stderr, "address mode: %i address: %i\n", am, vv[0]);
+#endif
                                    er=E_OVERFLOW;
+				}
                               else {
                                    t[1]=vv[0]&255;
                                    t[2]=(vv[0]>>8)&255;
@@ -1143,8 +1267,10 @@ int b_term(char *s, int *v, int *l, int pc)
      return(er);
 }
      
+/* translate a string into a first-pass sequence of tokens */
 static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
              int *na1, int *na2, int af, int *bytep)  /* Pass1 von s nach t */
+/* tr. pass1, from s to t */
 {
      static int v,f;
      static int operand,o;
@@ -1153,16 +1279,22 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
      int m, uz, byte;
      static unsigned char cast;
 
-     *nk=0;         /* Anzahl Komma               */
-     *na1=0;        /* Anzahl "asc-texte"         */
-     *na2=0;        /* und Anzahl Byte in diesen  */
+/* ich verstehe deutsch, aber verstehen andere leute nicht; so, werde ich
+   diese bemerkungen uebersetzen ... cameron */
+/* I understand German, but other folks don't, so I'll translate these
+   comments ... Cameron */
+/* note that I don't write so good tho' ;) */
+
+     *nk=0;         /* Anzahl Komma               *//* comma count */
+     *na1=0;        /* Anzahl "asc-texte"         *//* asc text count */
+     *na2=0;        /* und Anzahl Byte in diesen  *//* sum total bytecount */
      ll=0;
-     er=E_OK;
+     er=E_OK;		/* error state */
      p=0;
      q=0;
      ud = uz = byte =0;
-     mk=0;          /* 0 = mehrere Kommas erlaubt */
-     fl=0;          /* 1 = text einfach weitergeben */
+     mk=0;          /* 0 = mehrere Kommas erlaubt *//* 0 = add'l commas ok */
+     fl=0;          /* 1 = text einfach weitergeben *//* 1 = pass text thru */
      afl=0;         /* pointer flag for label */
 
      while(s[p]==' ') p++;
@@ -1175,12 +1307,15 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
           while(s[p]!='\0' && s[p]!=';')
           {
 
+		/* is keyword? */
                if(!(er=t_keyword(s+p,&ll,&n)))
                     break;
 
+		/* valid syntax, but just not a real token? */
                if(er && er!=E_NOKEY)
                     break;
 
+		/* if so, try to understand as label */
                if((er=l_def((char*)s+p,&ll,&n,&f)))
                     break;
 
@@ -1208,14 +1343,13 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
                     ll=n=0;
                     break;
                } else
-               if(s[p]==':')
+               if(s[p]==':')	/* to support label: ... syntax */
                {
                     p++;
                     while(s[p]==' ') p++;
                     l_set(n,pc,segment);        /* set as address value */
                     n=0;
-               } else
-               {
+               } else {		/* label ... syntax */
                     l_set(n,pc,segment);        /* set as address value */
                     n=0;
                }
@@ -1223,7 +1357,7 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
           }
 
           if((n & 0xff) <=Lastbef)
-               mk=1;     /* 1= nur 1 Komma erlaubt     */
+               mk=1;     /* 1= nur 1 Komma erlaubt *//* = only 1 comma ok */
      }
 
      if(s[p]=='\0' || s[p]==';')
@@ -1255,22 +1389,19 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
                while(s[p]==' ') p++;
           }
 
+/*
+ *
+ * operand processing
+ * byte = length of operand in bytes to be assembled
+ *
+ *
+ */
+
           /* this addresses forced high/low/two byte addressing, but only
              for the first operand. Further processing is done in a_term()
           */
-          if(s[p]=='<' || s[p]=='>') {
-              byte=1;
-              t[q++]=s[p++];
-              while(s[p]==' ') p++;
-          } else
-          if(s[p]=='!') {
-              byte=2;
-/*
-              t[q++]=s[p++];
-*/
-	      p++;
-              while(s[p]==' ') p++;
-          }
+
+/* FIXIT2 */
 
           while(s[p]!='\0' && s[p]!=';' && !er)
           {
@@ -1281,25 +1412,30 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
                {
                  if(operand)
                  {
-                    if(s[p]=='!' || s[p]=='@')
+			/* are we forcing the operand into a particular
+				addressing mode? !, @, ` operators */
+                    if(s[p]=='!' || s[p]=='@' || s[p]=='`')
                     {
                        cast=s[p];
                        operand= -operand+1;
                        p++;
                     } else
-                    if(s[p]=='(' || s[p]=='-' || s[p]=='>' || s[p]=='<' || s[p]=='[')
+                    if(s[p]=='(' || s[p]=='-' || s[p]=='>' ||
+			s[p]=='<' || s[p]=='[')
                     {
                          t[q++]=s[p++];
-                         operand= -operand+1;   /* invert to become reinverted */
+                         operand= -operand+1; /* invert to become reinverted */
                     } else
                     if(s[p]=='*')
                     {
                          t[q++]=s[p++];
                     } else
+		/* maybe it's a label */
                     if(isalpha(s[p]) || s[p]=='_')
                     {
                          m=n;
-                         er=l_such((char*)s+p,&ll,&n,&v,&afl);
+                         er=l_search((char*)s+p,&ll,&n,&v,&afl);
+
 /*
                          if(m==Kglobl || m==Kextzero) {
                               if(er==E_NODEF) {
@@ -1310,6 +1446,7 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
                               t[q++]=(n>>8) & 255;
                          } else
 */
+
                          if(!er)
                          {
                            if(afl) {
@@ -1323,6 +1460,9 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
                          } else
                          if(er==E_NODEF)
                          {
+#ifdef DEBUG_AM
+fprintf(stderr, "could not find %s\n", (char *)s+p);
+#endif
                               t[q++]=T_LABEL;
                               t[q++]=n & 255;
                               t[q++]=(n>>8) & 255;
@@ -1342,6 +1482,8 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
                          wval(q,v);
                     }
                     else
+
+		/* handle encodings: hex, binary, octal, quoted strings */
                     switch(s[p]) {
                     case '$':
                          tg_hex(s+p+1,&ll,&v);
@@ -1478,7 +1620,9 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
                     {
                          t[q++]=o;
                          p+=lp[o];
-                         uz++;  /* disables 8bit detection */
+#if(0)
+                         uz++; /* disable 8-bit detection */
+#endif
                     }
                     operand= -operand+1;
                  }
@@ -1489,34 +1633,34 @@ static int t_conv(signed char *s, signed char *t, int *l, int pc, int *nk,
      }
      if(!er)
      {
+/*
           if(uz==1 && ud==1 && byte!=2) {
                byte=1;
 	  }
-/*
           if(byte == 1) {
                t[q++] = T_FBYTE;
           } else if(byte == 2) {
                t[q++] = T_FADDR;
           }
 */
+	  byte = 0;
           t[q++]=T_END;
-          if(ud) {
+          if(ud > 0) {
                er=E_NODEF;
+		byte = 1;
           }
      }
-     *l=q;
-
-     /* FIXME: this is a of two "!" implementations :-( */
+     /* FIXME: this is an unholy union of two "!" implementations :-( */
      t[q++]='\0';
      t[q++]=cast;
-     if(bytep) *bytep=byte;
-
+     *l=q;
+     if(bytep) *bytep=byte; 
      return(er);
 }
 
 static int t_keyword(signed char *s, int *l, int *n)
 {
-     int i=0,j,hash;
+     int i = 0, j = 0, hash;
 
      if(!isalpha(s[0]) && s[0]!='.' && s[0]!='*' )
           return(E_NOKEY);
